@@ -1,18 +1,20 @@
-package CPAN::Mini;
-our $VERSION = '0.37_02';
-
 use strict;
 use warnings;
 
+package CPAN::Mini;
+our $VERSION = '0.560_01';
+
+## no critic RequireCarping
+
 =head1 NAME
 
 CPAN::Mini - create a minimal mirror of CPAN
 
 =head1 VERSION
 
-version 0.37_02
+version 0.560_01
 
- $Id: /my/rjbs/code/minicpan/trunk/lib/CPAN/Mini.pm 14501 2005-09-06T04:24:37.553688Z rjbs  $
+ $Id: /my/cs/projects/CPAN-Mini/trunk/lib/CPAN/Mini.pm 31471 2007-04-25T12:45:07.318475Z rjbs  $
 
 =head1 SYNOPSIS
 
@@ -47,21 +49,23 @@ the newest version of every distribution.  Those files are:
 
 =cut
 
-use Carp;
+use Carp ();
 
-use File::Path qw(mkpath);
-use File::Basename qw(basename dirname);
-use File::Spec::Functions qw(catfile canonpath);
-use File::Find qw(find);
+use File::Basename ();
+use File::Copy ();
+use File::Find ();
+use File::Path ();
+use File::Spec ();
+use File::Temp ();
 
 use URI ();
-use LWP::Simple qw(mirror RC_OK RC_NOT_MODIFIED);
+use LWP::Simple ();
 
-use Compress::Zlib qw(gzopen $gzerrno);
+use Compress::Zlib ();
 
 =head1 METHODS
 
-=head2 C<< update_mirror( %args ) >>
+=head2 update_mirror
 
  CPAN::Mini->update_mirror(
    remote => "http://cpan.mirrors.comintern.su",
@@ -88,6 +92,10 @@ The following options are recognized:
 Generally an octal number, this option sets the permissions of created
 directories.  It defaults to 0711.
 
+=item * C<exact_mirror>
+
+If true, the C<files_allowed> method will allow all extra files to be mirrored.
+
 =item * C<force>
 
 If true, this option will cause CPAN::Mini to read the entire module list and
@@ -122,7 +130,7 @@ setting would skip all distributions from RJBS and SUNGO:
 
 =item * C<module_filters>
 
-This options provides a set of rules for filtering modules.  It behaves like
+This option provides a set of rules for filtering modules.  It behaves like
 path_filters, but acts only on module names.  (Since most modules are in
 distributions with more than one module, this setting will probably be less
 useful than C<path_filters>.)  For example, this setting will skip any
@@ -130,6 +138,11 @@ distribution containing only modules with the word "Acme" in them:
 
  module_filters => [ qr/Acme/i ]
 
+=item * C<also_mirror>
+
+This option should be an arrayref of extra files in the remote CPAN to mirror
+locally.
+
 =item * C<skip_cleanup>
 
 If this option is true, CPAN::Mini will not try delete unmirrored files when it
@@ -146,12 +159,20 @@ sub update_mirror {
 	# mirrored tracks the already done, keyed by filename
 	# 1 = local-checked, 2 = remote-mirrored
 	$self->mirror_indices;
-	
+
 	return unless $self->{force} or $self->{changes_made};
 
+  $self->_mirror_extras;
+
 	# now walk the packages list
-	my $details = catfile($self->{local}, qw(modules 02packages.details.txt.gz));
-	my $gz = gzopen($details, "rb") or die "Cannot open details: $gzerrno";
+	my $details = File::Spec->catfile(
+    $self->{scratch},
+    qw(modules 02packages.details.txt.gz)
+  );
+
+	my $gz = Compress::Zlib::gzopen($details, "rb")
+    or die "Cannot open details: $Compress::Zlib::gzerrno";
+
 	my $inheader = 1;
 	while ($gz->gzreadline($_) > 0) {
 		if ($inheader) {
@@ -169,12 +190,16 @@ sub update_mirror {
 		$self->mirror_file("authors/id/$path", 1);
 	}
 
+  $self->_install_indices;
+
 	# eliminate files we don't need
 	$self->clean_unmirrored unless $self->{skip_cleanup};
 	return $self->{changes_made};
 }
 
-=head2 C<< new >>
+=head2 new
+
+  my $minicpan = CPAN::Mini->new;
 
 This method constructs a new CPAN::Mini object.  Its parameters are described
 above, under C<update_mirror>.
@@ -185,24 +210,38 @@ sub new {
 	my $class = shift;
 	my %defaults = (
     changes_made => 0,
-    dirmode      => 0711,
+    dirmode      => 0711, ## no critic Zero
     errors       => 1,
     mirrored     => {}
   );
 
 	my $self = bless { %defaults, @_ } => $class;
 
-	croak "no local mirror supplied"  unless $self->{local};
-  croak "no write permission to local mirror" unless -w $self->{local};
+  $self->{scratch} ||= File::Temp::tempdir(CLEANUP => 1);
 
-	croak "no remote mirror supplied" unless $self->{remote};
-  croak "unable to contact the remote mirror"
+	Carp::croak "no local mirror supplied"  unless $self->{local};
+
+  substr($self->{local}, 0, 1, $class->__homedir)
+    if substr($self->{local}, 0, 1) eq q{~};
+
+  Carp::croak "local mirror path exists but is not a directory"
+    if (-e $self->{local}) and not (-d $self->{local});
+
+  File::Path::mkpath($self->{local}, $self->{trace}, $self->{dirmode})
+    unless -e $self->{local};
+
+  Carp::croak "no write permission to local mirror" unless -w $self->{local};
+
+	Carp::croak "no remote mirror supplied" unless $self->{remote};
+  Carp::croak "unable to contact the remote mirror"
     unless LWP::Simple::head($self->{remote});
 
 	return $self;
 }
 
-=head2 C<< mirror_indices >>
+=head2 mirror_indices
+
+  $minicpan->mirror_indices;
 
 This method updates the index files from the CPAN.
 
@@ -211,14 +250,61 @@ This method updates the index files from the CPAN.
 sub mirror_indices {
 	my $self = shift;
 
-	$self->mirror_file($_) for qw(
-	                              authors/01mailrc.txt.gz
-	                              modules/02packages.details.txt.gz
-	                              modules/03modlist.data.gz
-	                             );
+  my @fixed_mirrors = qw(
+    authors/01mailrc.txt.gz
+    modules/02packages.details.txt.gz
+    modules/03modlist.data.gz
+  );
+
+  File::Path::mkpath(File::Spec->catdir($self->{scratch}, $_))
+    for qw(authors modules);
+
+  for my $path (@fixed_mirrors, @{$self->{also_mirror}}) {
+    my $local_file   = File::Spec->catfile($self->{local}, split m{/}, $path);
+    my $scratch_file = File::Spec->catfile($self->{scratch}, split m{/}, $path);
+
+    File::Copy::copy($local_file, $scratch_file);
+
+    utime((stat $local_file)[8,9], $scratch_file);
+
+    $self->mirror_file($path, undef, { to_scratch => 1 });
+  }
 }
 
-=head2 C<< mirror_file($path, $skip_if_present) >>
+sub _mirror_extras {
+	my $self = shift;
+
+  for my $path (@{$self->{also_mirror}}) {
+    $self->mirror_file($path, undef);
+  }
+}
+
+sub _install_indices {
+	my $self = shift;
+
+  my @fixed_mirrors = qw(
+    authors/01mailrc.txt.gz
+    modules/02packages.details.txt.gz
+    modules/03modlist.data.gz
+  );
+
+  for my $path (@fixed_mirrors) {
+    my $local_file = File::Spec->catfile($self->{local}, split m{/}, $path);
+
+    unlink $local_file;
+
+    File::Copy::copy(
+      File::Spec->catfile($self->{scratch}, split m{/}, $path),
+      File::Spec->catfile($self->{local},   split m{/}, $path),
+    );
+
+		$self->{mirrored}{$local_file} = 1;
+  }
+}
+
+=head2 mirror_file
+
+  $minicpan->mirror_file($path, $skip_if_present)
 
 This method will mirror the given file from the remote to the local mirror,
 overwriting any existing file unless C<$skip_if_present> is true.
@@ -226,15 +312,20 @@ overwriting any existing file unless C<$skip_if_present> is true.
 =cut
 
 sub mirror_file {
-	my $self   = shift;
-	my $path   = shift;           # partial URL
-	my $skip_if_present = shift;  # true/false
+  my ($self, $path, $skip_if_present, $arg) = @_;
+
+  $arg ||= {};
 
   # full URL
-	my $remote_uri = URI->new_abs($path, $self->{remote})->as_string;
+	my $remote_uri = eval { $path->isa('URI') }
+                 ? $path
+                 : URI->new_abs($path, $self->{remote})->as_string;
 
   # native absolute file
-	my $local_file = catfile($self->{local}, split "/", $path);
+	my $local_file = File::Spec->catfile(
+    $arg->{to_scratch} ? $self->{scratch} : $self->{local},
+    split m{/}, $path
+  );
 
 	my $checksum_might_be_up_to_date = 1;
 
@@ -245,16 +336,22 @@ sub mirror_file {
 		## upgrade to full mirror
 		$self->{mirrored}{$local_file} = 2;
 
-		mkpath(dirname($local_file), $self->{trace}, $self->{dirmode});
-		$self->trace($path);
-		my $status = mirror($remote_uri, $local_file);
+		File::Path::mkpath(
+      File::Basename::dirname($local_file),
+      $self->{trace},
+      $self->{dirmode}
+    );
 
-		if ($status == RC_OK) {
+		$self->trace($path);
+		my $status = LWP::Simple::mirror($remote_uri, $local_file);
+
+		if ($status == LWP::Simple::RC_OK) {
+      utime undef, undef, $local_file if $arg->{update_times};
 			$checksum_might_be_up_to_date = 0;
 			$self->trace(" ... updated\n");
 			$self->{changes_made}++;
-		} elsif ($status != RC_NOT_MODIFIED) {
-			warn( ($self->{trace} ? "\n" : '')
+		} elsif ($status != LWP::Simple::RC_NOT_MODIFIED) {
+			warn( ($self->{trace} ? "\n" : q{})
         . "$remote_uri: $status\n") if $self->{errors};
 			return;
 		} else {
@@ -264,7 +361,7 @@ sub mirror_file {
 
 	if ($path =~ m{^authors/id}) { # maybe fetch CHECKSUMS
 		my $checksum_path =
-			URI->new_abs("CHECKSUMS", $remote_uri)->rel($self->{remote});
+			URI->new_abs("CHECKSUMS", $remote_uri)->rel($self->{remote})->as_string;
 		if ($path ne $checksum_path) {
 			$self->mirror_file($checksum_path, $checksum_might_be_up_to_date);
 		}
@@ -273,15 +370,17 @@ sub mirror_file {
 
 =begin devel
 
-=head2 C<< _filter_module({ module => $foo, version => $foo, path => $foo }) >>
+=head2 _filter_module
 
-This internal-only method encapsulates the logic where we figure out if a
-module is to be mirrored or not. Better stated, this method holds the filter
-chain logic. C<update_mirror()> takes an optional set of filter parameters.  As
-C<update_mirror()> encounters a distribution, it calls this method to figure
-out whether or not it should be downloaded. The user provided filters are taken
-into account. Returns 1 if the distribution is filtered (to be skipped).
-Returns 0 if the distribution is to not filtered (not to be skipped).
+ next if
+   $self->_filter_module({ module => $foo, version => $foo, path => $foo });
+
+This method holds the filter chain logic. C<update_mirror> takes an optional
+set of filter parameters.  As C<update_mirror> encounters a distribution, it
+calls this method to figure out whether or not it should be downloaded. The
+user provided filters are taken into account. Returns 1 if the distribution is
+filtered (to be skipped).  Returns 0 if the distribution is to not filtered
+(not to be skipped).
 
 =end devel
 
@@ -305,7 +404,7 @@ sub __do_filter {
 sub _filter_module {
 	my $self = shift;
 	my $args = shift;
- 
+
 	if ($self->{skip_perl}) {
 		return 1 if $args->{path} =~ m{/(?:emb|syb|bio)*perl-\d}i;
 		return 1 if $args->{path} =~ m{/(?:parrot|ponie)-\d}i;
@@ -317,70 +416,132 @@ sub _filter_module {
 	return 0;
 }
 
-=head2 C<< file_allowed($file) >>
+=head2 file_allowed
+
+  next unless $minicpan->file_allowed($filename);
 
 This method returns true if the given file is allowed to exist in the local
 mirror, even if it isn't one of the required mirror files.
 
-By default, only dot-files are allowed.
+By default, only dot-files are allowed.  If the C<exact_mirror> option is true,
+all files are allowed.
 
 =cut
 
 sub file_allowed {
 	my ($self, $file) = @_;
 	return if $self->{exact_mirror};
-	return (substr(basename($file),0,1) eq '.') ? 1 : 0;
+	return (substr(File::Basename::basename($file),0,1) eq q{.}) ? 1 : 0;
 }
 
-=head2 C<< clean_unmirrored >>
+=head2 clean_unmirrored
 
-This method finds any files in the local mirror which are no longer needed and
-calls the C<clean_file> method on them.
+  $minicpan->clean_unmirrored;
+
+This method looks through the local mirror's files.  If it finds a file that
+neither belongs in the mirror nor is allowed (see the C<file_allowed> method),
+C<clean_file> is called on the file.
 
 =cut
 
 sub clean_unmirrored {
 	my $self = shift;
 
-	find sub {
-		my $file = canonpath($File::Find::name);
-		$self->clean_file($file);
+	File::Find::find sub {
+		my $file = File::Spec->canonpath($File::Find::name); ## no critic Package
+    return unless (-f $file and not $self->{mirrored}{$file});
+    return if $self->file_allowed($file);
+    $self->trace("cleaning $file ...");
+		if ($self->clean_file($file)) {
+      $self->trace("done\n");
+    } else {
+      $self->trace("couldn't be cleaned\n");
+    }
 	}, $self->{local};
 }
 
-=head2 C<< clean_file($filename) >>
+=head2 clean_file
 
-This method, called by C<clean_unmirrored>, checks whether the named file
-exists.  If it exists, and was not mirrored, and C<file_allowed> doesn't say it
-can stay, the file is deleted.
+  $minicpan->clean_file($filename);
+
+This method, called by C<clean_unmirrored>, deletes the named file.  It returns
+true if the file is successfully unlinked.  Otherwise, it returns false.
 
 =cut
 
 sub clean_file {
 	my ($self, $file) = @_;
 
-	return unless (-f $file and not $self->{mirrored}{$file});
-	return if $self->file_allowed($file);
 	unless (unlink $file) {
-    warn "$file ... cannot be removed: $!" if $self->{errors};
+    warn "$file ... cannot be removed: $!\n" if $self->{errors};
     return;
   }
-	$self->trace("$file ... removed\n");
+  return 1;
 }
 
-=head2 C<< trace( $message, $force ) >>
+=head2 trace
+
+  $minicpan->trace($message);
 
 If the object is mirroring verbosely, this method will print messages sent to
-it.  If CPAN::Mini is not operating in verbose mode, but C<$force> is true, it
-will print the message anyway.
+it.
 
 =cut
 
 sub trace {
-	my ($self, $message, $force) = @_;
-	print "$message" if $self->{trace} or $force;
+	my ($self, $message) = @_;
+	print "$message" if $self->{trace};
 }
 
+=head2 read_config
+
+  my %config = CPAN::Mini->read_config;
+
+This routine returns a set of arguments that can be passed to CPAN::Mini's
+C<new> or C<update_mirror> methods.  It will look for a file called
+F<.minicpanrc> in the user's home directory as determined by
+L<File::HomeDir|File::HomeDir>.
+
+=cut
+
+sub __homedir {
+  my ($class) = @_;
+
+  my $homedir = File::HomeDir->my_home || $ENV{HOME};
+
+  Carp::croak "couldn't determine your home directory!  set HOME env variable"
+    unless defined $homedir;
+  
+  return $homedir;
+}
+
+sub read_config {
+  my ($class) = @_;
+
+  my $filename = File::Spec->catfile($class->__homedir, '.minicpanrc');
+
+  return unless -e $filename;
+
+  open my $config_file, '<', $filename
+    or die "couldn't open config file $filename: $!";
+  
+  my %config;
+  while (<$config_file>) { 
+    chomp;
+    next if /\A\s*\Z/sm;
+    if (/\A(\w+):\s*(.+)\Z/sm) { $config{$1} = $2; }
+  }
+  for (qw(also_mirror)) {
+    $config{$_} = [ grep { length } split /\s+/, $config{$_}] if $config{$_};
+  }
+  for (qw(module_filters path_filters)) {
+    $config{$_} = [ map { qr/$_/ } split /\s+/, $config{$_} ] if $config{$_};
+  }
+  return %config;
+}
+
+=head2 
+
 =head1 SEE ALSO
 
 Randal Schwartz's original article on minicpan, here:
@@ -401,11 +562,19 @@ Thanks to Shawn Sorichetti for fixing a stupid octal-number-as-string bug.
 Thanks to sungo for implementing the filters, so I can finally stop mirroring
 bioperl, and Robert Rothenberg for suggesting adding coderef rules.
 
+Thanks to Adam Kennedy for noticing and complaining about a lot of stupid
+little design decisions.
+
+Thanks to Michael Schwern and Jason Kohles, for pointing out missing
+documentation.
+
 =head1 AUTHORS
 
-Randal Schwartz <F<merlyn@stonehenge.com>> did all the work. 
+Randal Schwartz <F<merlyn@stonehenge.com>> wrote the original F<minicpan>
+script.
 
-Ricardo SIGNES <F<rjbs@cpan.org>> made a module and distribution.
+Ricardo SIGNES <F<rjbs@cpan.org>> turned Randal's script into a module and CPAN
+distribution, and has maintained it since its release as such.
 
 This code was copyrighted in 2004, and is released under the same terms as Perl
 itself.
